require "ostruct"
module Gitvers
  class NoVersionError < StandardError
  end
  class InvalidVersionError < StandardError
  end

  class Repository
    def initialize(path, commit: 'HEAD')
      @path = path
      @commit_name = commit

      while path != '/'
        git_dir = File.join(path, "/.git")
        head_file = File.join(path, "/HEAD")
        if File.exist?(git_dir)
          @git_dir = git_dir
          break
        elsif File.exists?(head_file)
          @git_dir = path
          break
        end
        path = File.expand_path(File.join(path, '..'))
      end
      @git_dir or raise "Invalid git repository #{@path}"
    end

    def summary
      "#{full_version} (#{revision_number})"
    end

    def full_version
      check!
      git(:describe).strip.gsub(/^v/, '')
    end

    def short_version
      check!
      git('describe --abbrev=0').strip.gsub(/^v/, '')
    end

    def revision_number
      check!
      git("rev-list #{@commit_name}").lines.count
    end

    def versions
      return @versions if defined?(@versions)
      @versions = git(:tag).lines.reverse
    end

    def bump(version)
      check!

      ver = check_version!(version) rescue nil

      if ver.nil?
        ver = check_version!(short_version)
        case version.to_sym
        when :major
          ver.major += 1
          ver.minor = ver.patch = 0
        when :minor
          ver.minor += 1
          ver.patch = 0
        when :patch
          ver.patch += 1
        end
      end
      tag("#{ver.major}.#{ver.minor}.#{ver.patch}")
    end

    def tag(version)
      check_version!(version)
      git("tag -a v#{version} -m v#{version}")
      puts "Created tag #{version}. You need to push it with git push --tags"
    end

    def check!
      versions.count > 0 or raise NoVersionError
    end

    def check_version!(version)
      version =~ /(\d+)\.(\d+)\.(\d+)/ or raise InvalidVersionError, "Invalid version number #{version || '(null)'}"

      OpenStruct.new({
        :major => $1.to_i,
        :minor => $2.to_i,
        :patch => $3.to_i
      })
    end

    private
    def git(command)
      `git --git-dir="#{@git_dir}" --work-tree="#{@path}/" #{command}`
    end
  end
end
